#!/usr/bin/env node

'use strict';

var commander = require('commander');
var path = require('path');
var rxjs = require('rxjs');
var operators = require('rxjs/operators');
var fs = require('fs');
var inquirer = require('inquirer');

function _interopNamespace(e) {
    if (e && e.__esModule) return e;
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () {
                        return e[k];
                    }
                });
            }
        });
    }
    n['default'] = e;
    return Object.freeze(n);
}

var path__namespace = /*#__PURE__*/_interopNamespace(path);
var fs__namespace = /*#__PURE__*/_interopNamespace(fs);
var inquirer__namespace = /*#__PURE__*/_interopNamespace(inquirer);

const constants = {
    rest: {
        baseUrl: 'https://scrapbox.io/api',
    },
    websocket: {
        endpoint: 'wss://scrapbox.io/socket.io/?EIO=4&transport=websocket',
        origin: 'https://scrapbox.io',
        responseTimeout: 1000 * 15,
        packetTypes: {
            initialize: '0',
            ping: '2',
            pong: '3',
            connected: '40',
            // e.g.
            //   * 42: 'cursor'
            //   * 42X: user custom request (X is arbitrary natural number to specify response)
            send: '42',
            // e.g.
            //   * 43X: response to user custom request (X is number used for request)
            receive: '43',
            response: '43',
        },
    },
};

// copied from official index.js
// modified original toString16 of official index.js
const toString16 = (userIdOrNumber, length) => {
    const hexOrId = userIdOrNumber.toString(16);
    return hexOrId.length > length ? hexOrId.slice(hexOrId.length - length) : '0'.repeat(length - hexOrId.length) + hexOrId;
};
const map = new Map();
const getIdGenerator = (userId) => {
    if (map.has(userId)) {
        return map.get(userId);
    }
    let counter = Math.floor(16777215 * Math.random());
    const generate = () => {
        const t = Math.floor(Date.now() / 1e3);
        if ((counter += 1) > 16777215) {
            counter = 0;
        }
        return (toString16(t, 8) + toString16(userId, 6) + toString16(0, 4) + toString16(counter, 6));
    };
    map.set(userId, generate);
    return generate;
};
const generateId = (userId) => {
    const generator = getIdGenerator(userId);
    return generator();
};

const isBrowser = () => typeof window !== 'undefined';

// NOTE: use bind(window) to avoid `TypeError: Illegal invocation`
const fetch = isBrowser() ? globalThis.fetch.bind(window) : require('node-fetch');
class DefaultHttpClient {
    fetchFn;
    constructor(fetchFn = fetch) {
        this.fetchFn = fetchFn;
    }
    async get(url, options = {}) {
        const result = await this.fetchFn(url, options);
        return (await result.json());
    }
}

const isObject = (data) => typeof data === 'object' && data !== null;
const isNotLoggedInError = (data) => isObject(data) && data['name'] === 'NotLoggedInError';
class RestApi {
    token;
    httpClient;
    constructor(token, httpClient = new DefaultHttpClient()) {
        this.token = token;
        this.httpClient = httpClient;
    }
    async getPage(projectName, pageName) {
        return this.request(`${constants.rest.baseUrl}/pages/${encodeURIComponent(projectName)}/${encodeURIComponent(pageName)}?followRename=true`);
    }
    async getProject(projectName) {
        return this.request(`${constants.rest.baseUrl}/projects/${encodeURIComponent(projectName)}`);
    }
    async getProjects() {
        return this.request(`${constants.rest.baseUrl}/projects`);
    }
    async getMe() {
        return this.request(`${constants.rest.baseUrl}/users/me`);
    }
    async request(url) {
        const result = await this.httpClient.get(url, {
            headers: {
                Cookie: `connect.sid=${this.token}`,
            },
        });
        if (isNotLoggedInError(result)) {
            throw new Error('NotLoggedInError');
        }
        return result;
    }
}

const isIntegerString = (value) => Number.isInteger(Number.parseInt(value));
const getAuthCookieValue = (token) => `connect.sid=${token}`;

const toSocketIoMessagePayload = (id, data) => {
    return `${constants.websocket.packetTypes.send}${id}${JSON.stringify(['socket.io-request', data])}`;
};
const isConnectionMessage = (message) => {
    return message[0] === constants.websocket.packetTypes.initialize;
};
const isResponseMessage = (message) => {
    return message[0].length > 2 && message[0].startsWith(constants.websocket.packetTypes.response);
};
const getRequestId = (packetType) => packetType.slice(2);
const isResponseMessageOf = (sid) => (message) => {
    return isResponseMessage(message) && sid === getRequestId(message[0]);
};
const isResponseOf = (expected) => {
    return ([sid]) => expected === sid;
};
// 430[{...}}] => 430, [{}]
// @see https://github.com/socketio/engine.io-protocol
const scrapboxDeserializer = (event) => {
    let packetType = '';
    let message = String(event.data);
    while (message.length) {
        const head = message[0];
        if (isIntegerString(head)) {
            packetType += head;
            message = message.substr(1);
        }
        else {
            return [packetType, JSON.parse(message)];
        }
    }
    return [packetType, undefined];
};

// NOTE: userId is used to generate new ID
const createChanges = (params, userId) => params.map((param) => {
    switch (param.type) {
        case 'insert': {
            return { _insert: param.position || '_end', lines: { id: generateId(userId), text: param.text } };
        }
        case 'update': {
            return { _update: param.id, lines: { text: param.text } };
        }
        case 'delete': {
            return { _delete: param.id, lines: -1 };
        }
        case 'title': {
            return { title: param.title };
        }
        case 'description': {
            return { descriptions: [param.text] };
        }
        default: {
            throw new Error('Bad implement');
        }
    }
});

const time = () => new Date().getTime();
const debugWebsocket = (websocket) => {
    console.log(time(), '[websocket] initialized');
    websocket.addEventListener('open', () => console.log(time(), '[websocket] onopen'));
    websocket.addEventListener('message', (ev) => console.log(time(), '[websocket] onmessage', ev.data));
    websocket.addEventListener('error', (ev) => console.log(time(), '[websocket] onerror'));
    websocket.addEventListener('close', (ev) => console.log(time(), '[websocket] onclose', ev.reason));
    websocket.send = new Proxy(websocket.send, {
        apply: (target, thisArg, args) => {
            console.log(time(), '[websocket] send', args[0]);
            return target.call(websocket, args[0]);
        },
    });
    websocket.close = new Proxy(websocket.close, {
        apply: (target) => {
            console.log(time(), '[websocket] close');
            return target.call(websocket);
        },
    });
};

const defaultDeserializer = (message) => message.data;
class RxWebSocket extends rxjs.Subject {
    #websocket;
    #open$;
    constructor(url, { protocols, clientOptions, deserializer = defaultDeserializer, debug = false }) {
        super();
        const websocket = isBrowser() ? new WebSocket(url, protocols) : new (require('ws'))(url, protocols, clientOptions);
        if (debug) {
            debugWebsocket(websocket);
        }
        const open$ = rxjs.fromEvent(websocket, 'open').pipe(rxjs.take(1), rxjs.shareReplay());
        const error$ = rxjs.fromEvent(websocket, 'error').pipe(rxjs.take(1), rxjs.shareReplay());
        const close$ = rxjs.fromEvent(websocket, 'close').pipe(rxjs.take(1), operators.share());
        const message$ = rxjs.fromEvent(websocket, 'message').pipe(operators.takeUntil(close$), rxjs.shareReplay(1));
        rxjs.merge(open$.pipe(rxjs.ignoreElements()), message$.pipe(operators.map(deserializer)), error$.pipe(rxjs.mergeMap(rxjs.throwError)), close$.pipe(rxjs.ignoreElements())).subscribe(this);
        this.#open$ = open$;
        this.#websocket = websocket;
    }
    send(data) {
        this.#open$.subscribe(() => this.#websocket.send(data));
        return this;
    }
    close() {
        this.#websocket.close();
        this.unsubscribe();
    }
}

class ScrapboxWebsocketHandler {
    #room = null;
    #sid = 0;
    #socket$;
    #close$ = new rxjs.Subject();
    constructor({ token = '', debug }) {
        this.#socket$ = new rxjs.Observable((subscriber) => {
            const socket = new RxWebSocket(constants.websocket.endpoint, {
                clientOptions: { headers: { Origin: constants.websocket.origin, Cookie: getAuthCookieValue(token) } },
                deserializer: scrapboxDeserializer,
                debug,
            });
            socket
                .pipe(operators.takeUntil(this.#close$), rxjs.filter(isConnectionMessage), 
            // send ack
            rxjs.tap(() => {
                socket.send(constants.websocket.packetTypes.connected);
            }), 
            // set ping
            rxjs.tap(([, data]) => rxjs.interval(data.pingInterval)
                .pipe(operators.takeUntil(socket))
                .subscribe(() => socket.send(constants.websocket.packetTypes.ping))), rxjs.delayWhen(() => socket.pipe(operators.first(isResponseOf(constants.websocket.packetTypes.connected)))), rxjs.mapTo(socket), operators.timeout(constants.websocket.responseTimeout))
                .subscribe(subscriber);
            return () => socket.close();
        }).pipe(rxjs.shareReplay(1));
    }
    close() {
        this.#close$.next(undefined);
    }
    async commit({ projectId, userId, pageId, parentId, changes, }) {
        return rxjs.firstValueFrom(this.#join({ projectId, pageId }).pipe(rxjs.concatMap(() => {
            return this.#send({
                method: 'commit',
                data: { kind: 'page', userId, projectId, pageId, parentId, changes: createChanges(changes, userId), cursor: null, freeze: true },
            });
        })));
    }
    #join({ projectId, pageId }) {
        return this.#room?.projectId === projectId && this.#room?.pageId === pageId
            ? rxjs.of({})
            : this.#send({
                method: 'room:join',
                data: { pageId, projectId, projectUpdatesStream: false },
            }).pipe(rxjs.tap({
                complete: () => {
                    this.#room = { projectId, pageId };
                },
            }));
    }
    #send(data) {
        const sid = String(this.#sid++);
        return this.#socket$.pipe(rxjs.take(1), rxjs.mergeMap((socket) => socket.send(toSocketIoMessagePayload(sid, data)).pipe(operators.first(isResponseMessageOf(sid)), operators.timeout(constants.websocket.responseTimeout))));
    }
}

class ScrapboxApi {
    restApi;
    websocketHandler;
    constructor(token = '', debug = false, restApi = new RestApi(token), websocketHandler = new ScrapboxWebsocketHandler({ token, debug })) {
        this.restApi = restApi;
        this.websocketHandler = websocketHandler;
    }
    async getPage(projectName, pageName) {
        return this.restApi.getPage(projectName, pageName);
    }
    async changeLine(projectName, pageName, change) {
        const project = await this.getProject(projectName);
        const { pageId, commitId } = await this.getPageIdAndCommitId(projectName, pageName);
        await this.commit({
            changes: Array.isArray(change) ? change : [change],
            projectId: project.id,
            pageId: pageId,
            commitId: commitId,
        });
    }
    async getUser() {
        return await this.restApi.getMe();
    }
    closeConnection() {
        this.websocketHandler.close();
    }
    async getProject(projectName) {
        return await this.restApi.getProject(projectName);
    }
    async getPageIdAndCommitId(projectName, pageName) {
        const page = await this.restApi.getPage(projectName, pageName);
        return { pageId: page.id, commitId: page.commitId };
    }
    async commit(param) {
        const user = await this.getUser();
        return this.websocketHandler.commit({
            userId: user.id,
            projectId: param.projectId,
            pageId: param.pageId,
            parentId: param.commitId,
            changes: param.changes,
        });
    }
}

const loadFile = (path) => new Promise((resolve, reject) => fs__namespace.readFile(path, 'utf8', (err, data) => {
    if (err) {
        reject(err);
    }
    resolve(data);
}));
const getJSTString = () => new Date().toLocaleString('ja-JP', { timeZone: 'Asia/Tokyo' });
const versionComment = `/* deployed at: ${getJSTString()} */`;
// first space is needed for scrapbox code block
const loadSourceCode = async (file) => ` ${versionComment}${await loadFile(file)}`;

const findIndex = (searchString, lines) => {
    return lines.findIndex((line) => line.text.includes(searchString));
};
const findNextLineIdOrFail = (searchString, lines) => {
    const index = findIndex(searchString, lines);
    if (index === -1) {
        throw new Error('Line not found');
    }
    const nextLineOrNone = lines[index + 1];
    if (!nextLineOrNone) {
        throw new Error('Next line not found');
    }
    return nextLineOrNone.id;
};

const isNonEmptyString = (value) => typeof value === 'string' && value.length > 0;
const isCSSFile = (value) => value.endsWith('.css');
const isJSFile = (value) => value.endsWith('.js');
const validateDeployArguments = ({ project, page, sourceFilePath, token }) => {
    const errors = [];
    if (!isNonEmptyString(project)) {
        errors.push('project is required');
    }
    if (!isNonEmptyString(page)) {
        errors.push('page is required');
    }
    if (!isNonEmptyString(sourceFilePath)) {
        errors.push('filepath is required');
    }
    if (!(isCSSFile(sourceFilePath) || isJSFile(sourceFilePath))) {
        errors.push('file type is not css or js');
    }
    if (!isNonEmptyString(token)) {
        errors.push('token is required');
    }
    return errors;
};

const deploy = async ({ page, sourceFilePath, project, token, debug = false, }) => {
    const errors = validateDeployArguments({ token, sourceFilePath, project, page });
    if (errors.length) {
        throw new Error(`ParameterError: ${JSON.stringify(errors)}`);
    }
    const codeBlockLabel = isCSSFile(sourceFilePath) ? 'style.css' : 'script.js';
    const sourceCode = await loadSourceCode(sourceFilePath);
    const scrapboxClient = new ScrapboxApi(token, debug);
    const pageData = await scrapboxClient.getPage(project, page);
    // NOTE: assuming code-block exists already
    const lineId = findNextLineIdOrFail(codeBlockLabel, pageData.lines);
    // FIXME: bad type assertion
    await scrapboxClient.changeLine(project, page, { type: 'update', id: lineId, text: sourceCode });
    scrapboxClient.closeConnection();
};

const selectProject = async ({ client }) => {
    const projectOptions = await client.getProjects().then(({ projects }) => projects.map(({ name }) => name));
    const answers = await inquirer__namespace.prompt([{ type: 'list', name: 'project', message: 'select target project', choices: projectOptions }]);
    return answers.project;
};

const pkg = require(path__namespace.resolve(__dirname, '../../package.json'));
const program = new commander.Command().version(pkg.version);
program
    .command('deploy-userscript <js-file>')
    .description('update userscript in user page')
    .option('-p, --project <type>', 'project name')
    .option('-d, --debug', 'output extra debugging')
    .action(async (jsFile, options) => {
    const token = process.env.TOKEN ?? '';
    if (!token) {
        throw new Error('TOKEN not set');
    }
    const api = new RestApi(token);
    const selectedProject = options.project ?? (await selectProject({ client: api }));
    const userName = await api.getMe().then(({ name }) => name);
    await deploy({ token, project: selectedProject, page: userName, sourceFilePath: jsFile, debug: options.debug });
});
program
    .command('deploy-usercss <css-file>')
    .description('update usercss in settings page')
    .option('-p, --project <type>', 'project name')
    .option('-d, --debug', 'output extra debugging')
    .action(async (jsFile, options) => {
    const token = process.env.TOKEN ?? '';
    if (!token) {
        throw new Error('TOKEN not set');
    }
    const api = new RestApi(token);
    const selectedProject = options.project ?? (await selectProject({ client: api }));
    await deploy({ token, project: selectedProject, page: 'settings', sourceFilePath: jsFile, debug: options.debug });
});
program.parseAsync(process.argv).catch((e) => {
    console.error(e);
    process.exit(1);
});
//# sourceMappingURL=bin.js.map
