import { Subject, Observable, filter, tap, interval, delayWhen, mapTo, shareReplay, firstValueFrom, concatMap, of, take, mergeMap } from 'rxjs';
import { takeUntil, first, timeout } from 'rxjs/operators';
import { constants } from '../common/constants.js';
import { scrapboxDeserializer, isConnectionMessage, isResponseOf, toSocketIoMessagePayload, isResponseMessageOf } from './internal/message.js';
import { createChanges } from './internal/request.js';
import { getAuthCookieValue } from './internal/utils.js';
import { RxWebSocket } from './internal/websocket.js';

class ScrapboxWebsocketHandler {
    #room = null;
    #sid = 0;
    #socket$;
    #close$ = new Subject();
    constructor({ token = '', debug }) {
        this.#socket$ = new Observable((subscriber) => {
            const socket = new RxWebSocket(constants.websocket.endpoint, {
                clientOptions: { headers: { Origin: constants.websocket.origin, Cookie: getAuthCookieValue(token) } },
                deserializer: scrapboxDeserializer,
                debug,
            });
            socket
                .pipe(takeUntil(this.#close$), filter(isConnectionMessage), 
            // send ack
            tap(() => {
                socket.send(constants.websocket.packetTypes.connected);
            }), 
            // set ping
            tap(([, data]) => interval(data.pingInterval)
                .pipe(takeUntil(socket))
                .subscribe(() => socket.send(constants.websocket.packetTypes.ping))), delayWhen(() => socket.pipe(first(isResponseOf(constants.websocket.packetTypes.connected)))), mapTo(socket), timeout(constants.websocket.responseTimeout))
                .subscribe(subscriber);
            return () => socket.close();
        }).pipe(shareReplay(1));
    }
    close() {
        this.#close$.next(undefined);
    }
    async commit({ projectId, userId, pageId, parentId, changes, }) {
        return firstValueFrom(this.#join({ projectId, pageId }).pipe(concatMap(() => {
            return this.#send({
                method: 'commit',
                data: { kind: 'page', userId, projectId, pageId, parentId, changes: createChanges(changes, userId), cursor: null, freeze: true },
            });
        })));
    }
    #join({ projectId, pageId }) {
        return this.#room?.projectId === projectId && this.#room?.pageId === pageId
            ? of({})
            : this.#send({
                method: 'room:join',
                data: { pageId, projectId, projectUpdatesStream: false },
            }).pipe(tap({
                complete: () => {
                    this.#room = { projectId, pageId };
                },
            }));
    }
    #send(data) {
        const sid = String(this.#sid++);
        return this.#socket$.pipe(take(1), mergeMap((socket) => socket.send(toSocketIoMessagePayload(sid, data)).pipe(first(isResponseMessageOf(sid)), timeout(constants.websocket.responseTimeout))));
    }
}

export { ScrapboxWebsocketHandler };
//# sourceMappingURL=websocket-handler.js.map
